/*
 * Copyright (C) 2006-2009 Andrey Kholmanskih. All rights reserved.
 * 
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library.  If not, write to the 
 *
 * Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330,
 * Boston, MA  USA  02111-1307
 * 
 * 
 * Andrey Kholmanskih
 * support@jdbreport.com
 */
package and.util;

import java.awt.Point;
import java.awt.geom.Point2D;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.text.DecimalFormat;
import java.util.Calendar;
import java.util.Date;

/**
 * @version 1.3 03.05.2009
 * @author Andrey Kholmanskih
 * 
 */
public class Utilities {

	private static char decimalSeparator = '\0';

	/**
	 * Linear Bézier curves
	 * 
	 * @param t
	 * @param p
	 * @return
	 */
	public static float lineBezier(float t, float[] p) {
		return (float) ((1 - t) * p[0] + t * p[1]);

	}

	/**
	 * Quadratic Bézier curves
	 * 
	 * @param t
	 * @param p
	 * @return
	 */
	public static float quadBezier(float t, float[] p) {
		return (float) (sqr(1 - t) * p[0] + 2 * t * (1 - t) * p[1] + sqr(t)
				* p[2]);

	}

	/**
	 * Cubic Bézier curves
	 * 
	 * @param t
	 * @param p
	 * @return
	 */
	public static float cubicBezier(float t, float[] p) {
		return (float) (cubic(1 - t) * p[0] + 3 * t * sqr(1 - t) * p[1] + 3
				* sqr(t) * (1 - t) * p[2] + cubic(t) * p[3]);
	}

	public static double sqr(double v) {
		return v * v;
	}

	public static double cubic(double v) {
		return v * v * v;
	}

	public static double round(double value) {
		return round(value, 0);
	}

	public static double round(double value, int decimal) {
		long d = 1;
		int c = decimal < 0 ? -decimal : decimal;
		for (int i = 0; i < c; i++) {
			d = d * 10;
		}
		if (decimal >= 0) {
			return (double)Math.round(value * d) / d;
		} else {
			return Math.round(value / d) * d;
		}
	}

	public static float round(float value, int decimal) {
		int d = 1;
		int c = decimal < 0 ? -decimal : decimal;
		for (int i = 0; i < c; i++) {
			d = d * 10;
		}
		if (decimal >= 0) {
			int v = Math.round(value * d);
			return 1.0f * v / d;
		} else {
			return Math.round(value / d) * d;
		}
	}
	
	public static String roundStr(Double value, int decimal) {
		return String.format("%1." + decimal + "f", value);
	}

	public static char getDecimalSeparator() {
		if (decimalSeparator  == '\0') {
			DecimalFormat df = new DecimalFormat();
			decimalSeparator = df.getDecimalFormatSymbols().getDecimalSeparator();
		}
		return decimalSeparator;
	}
	
	public static float parseFloat(String s) {
		if (s == null) return 0;
		try {
			return Float.parseFloat(s);
		} catch (NumberFormatException e) {
			char ds = s.indexOf(',') >= 0 ? '.' : ',';
			char oldChar = ds == ',' ? '.' : ',';
			s = s.replace(oldChar, ds);
			return Float.parseFloat(s);
		}
	}

	public static double parseDouble(String s) {
		if (s == null) return 0;
		try {
			return Double.parseDouble(s);
		} catch (NumberFormatException e) {
			char ds = s.indexOf(',') >= 0 ? '.' : ',';
			char oldChar = ds == ',' ? '.' : ',';
			s = s.replace(oldChar, ds);
			return Double.parseDouble(s);
		}
	}

	public static int parseInt(String s) {
		if (s == null) return 0;
		try {
			return Integer.parseInt(s);
		} catch (NumberFormatException e) {
			return (int)Math.round(parseDouble(s));
		}
	}

	public static long parseLong(String s) {
		try {
			return Long.parseLong(s);
		} catch (NumberFormatException e) {
			return Math.round(parseDouble(s));
		}
	}
	
	public static URL getDirFromResource(URL res) throws MalformedURLException {
		String file = res.getFile();
		int i = file.indexOf('!');
		if (i > 0) {
			file = file.substring(0, i);
		}
		file = file.substring(0, file.lastIndexOf("/"));
		return new URL(file);
	}

	public static String html2Plain(String text) {
		StringBuffer result = new StringBuffer();
		for (int i = 0; i < text.length(); i++) {
			if (text.charAt(i) != '&' || i >= text.length() - 2) {
				result.append(text.charAt(i));
			} else {
				i++;
				if (text.charAt(i) != '#') {
					result.append('&');
					result.append(text.charAt(i));
					continue;
				}
				int b = ++i;
				while (text.charAt(i) != ';' && i < text.length()) {
					i++;
				}
				if (text.charAt(i) == ';') {
					try {
						int ch = Integer.parseInt(text.substring(b, i));
						result.append((char) ch);
					} catch (Exception e) {
						result.append(text.substring(b - 2, i));
					}
				} else {
					result.append(text.substring(b - 2));
				}

			}
		}
		return result.toString();
	}

	public static String truncate(String value, int length) {
		if (value != null && value.length() > length) {
			return value.substring(0, length);
		}
		return value;
	}

	public static String getFileExtension(File f) {
		if (f != null) {
			return getFileExtension(f.getName());
		}
		return null;
	}

	public static String getFileExtension(String filename) {
		int i = filename.lastIndexOf('.');
		if (i > 0 && i < filename.length() - 1) {
			return filename.substring(i + 1).toLowerCase();
		}
		return "";
	}

	public static String changeFileExtension(String fileName,
			String newExtension) {
		int i = fileName.lastIndexOf('.');
		if (i > 0 && i < fileName.length() - 1) {
			return fileName.substring(0, i) + newExtension;
		}
		return fileName + "." + newExtension;
	}

	public static String extractFileName(String filePath) {
		int i = filePath.lastIndexOf('/');
		if (i < 0)
			i = filePath.lastIndexOf('\\');
		if (i < 0)
			return filePath;
		return filePath.substring(i + 1);
	}

	public static String extractFilePath(String filePath) {
		int i = filePath.lastIndexOf('/');
		if (i < 0)
			i = filePath.lastIndexOf('\\');
		if (i < 0)
			return "";
		return filePath.substring(0, i + 1);
	}

	/**
	 * Fill array from stream
	 * 
	 * @param in
	 * @param b
	 * @throws IOException
	 * @deprecate use readBytes
	 */
	public static void read(InputStream in, byte[] b) throws IOException {
		readBytes(in, b);
	}

	/**
	 * Fill array from stream
	 * 
	 * @param in
	 * @param b
	 * @return count bytes reads from stream
	 * @throws IOException
	 */
	public static int readBytes(InputStream in, byte[] b) throws IOException {
		int l = b.length;
		int pos = 0;
		int n;
		do {
			n = in.read(b, pos, l);
			if (n > 0) {
				pos += n;
				l -= n;
			}
		} while (n > 0 && l > 0);
		return pos;
	}

	public static String fileMD5(File file) throws NoSuchAlgorithmException,
			IOException {
		byte[] buf;
		InputStream stream = new BufferedInputStream(new FileInputStream(file));
		try {
			buf = new byte[stream.available()];
			read(stream, buf);
		} finally {
			stream.close();
		}
		MessageDigest md = MessageDigest.getInstance("MD5");
		md.update(buf);
		MessageDigest md2 = null;
		try {
			md2 = (MessageDigest) md.clone();
		} catch (CloneNotSupportedException e) {
			throw new IOException("Clone not supported");
		}
		byte[] mdbytes = md2.digest();
		return md5ToString(mdbytes);
	}

	public static String printMD5(String s) throws NoSuchAlgorithmException {
		MessageDigest md = MessageDigest.getInstance("MD5");
		byte[] buf = s.getBytes();
		md.update(buf);
		MessageDigest md2 = null;
		try {
			md2 = (MessageDigest) md.clone();
		} catch (CloneNotSupportedException e) {
			throw new NoSuchAlgorithmException(e);
		}
		byte[] mdbytes = md2.digest();
		return md5ToString(mdbytes);
	}

	public static String md5ToString(byte[] mdbytes) {
		StringBuffer md5 = new StringBuffer();
		for (int i = 0; i < mdbytes.length; i++) {
			String c = Integer.toHexString(mdbytes[i]);
			if (c.length() == 1)
				md5.append('0');
			else if (c.length() > 2)
				c = c.substring(c.length() - 2);
			md5.append(c);
		}
		return md5.toString();
	}

	/**
	 * Calculates distance between points p1 and p2
	 * 
	 * @param p1
	 * @param p2
	 * @return
	 */
	public static double getDistance(Point p1, Point p2) {
		double dy = Math.abs(p1.y - p2.y);
		double dx = Math.abs(p1.x - p2.x);
		return Math.sqrt(dy * dy + dx * dx);
	}

	/**
	 * Returns quarter by date
	 * 
	 * @param date
	 * @return
	 */
	public static int getQuarter(Date date) {
		if (date != null) {
			Calendar c = Calendar.getInstance();
			c.setTime(date);
			int m = c.get(Calendar.MONTH);
			switch (m) {
			case Calendar.JANUARY:
			case Calendar.FEBRUARY:
			case Calendar.MARCH:
				return 1;
			case Calendar.APRIL:
			case Calendar.MAY:
			case Calendar.JUNE:
				return 2;
			case Calendar.JULY:
			case Calendar.AUGUST:
			case Calendar.SEPTEMBER:
				return 3;
			case Calendar.OCTOBER:
			case Calendar.NOVEMBER:
			case Calendar.DECEMBER:
				return 4;
			}
		}
		return 0;
	}

	public static Date getStartDateQuarter(int quarter, int year) {
		Calendar c = Calendar.getInstance();
		int month = 0;
		switch (quarter) {
		case 1:
			month = 1;
			break;
		case 2:
			month = 4;
			break;
		case 3:
			month = 7;
			break;
		case 4:
			month = 10;
			break;
		}
		if (month > 0) {
			c.set(year, month - 1, 1, 0, 0, 0);
			c.set(Calendar.MILLISECOND, 0);
			return c.getTime();
		}
		return null;
	}

	public static Date getEndDateQuarter(int quarter, int year) {
		Calendar c = Calendar.getInstance();
		int month = 0;
		int d = 0;
		switch (quarter) {
		case 1:
			month = 3;
			d = 31;
			break;
		case 2:
			month = 6;
			d = 30;
			break;
		case 3:
			month = 9;
			d = 30;
			break;
		case 4:
			month = 12;
			d = 31;
			break;
		}
		if (month > 0) {
			c.set(year, month - 1, d, 0, 0, 0);
			c.set(Calendar.MILLISECOND, 0);
			return c.getTime();
		}
		return null;
	}

	/**
	 * 
	 * @param year
	 * @param quarter 
	 */
	public static int daysOfQuarter(int year, int quarter) {
		if (quarter == 1) {
		 if (year % 4 == 0) return 91; else return 90;
		} else if (quarter == 2) {
			return 91;
		}
		return 92;
	}
	
	/**
	 * Formatting time 
	 * @param t
	 * @return
	 */
	public static String timeFormat(long t) {
		t = t / 1000;
		long h = t / 3600;
		long m = (t - (h * 3600)) / 60;
		long s = t - (h * 3600) - m * 60;
		long d = h / 24;
		h = h - d * 24;
		String result = "";
		if (d > 0) {
			result = d + " day ";
		}
		result += h + ":";
		if (m < 10) result += "0" + m + ":";
		else result += m + ":";
		if (s < 10) result += "0" + s;
		else result += s;
		return result;
	}
	
	/**
	 * Turn of a point round other point counter-clockwise
	 * 
	 * @param x0
	 * @param y0
	 * @param x
	 * @param y
	 * @param a
	 *            Angle of rotation in radians
	 * @return
	 */
	public static Point2D.Double rotatePoint(double x0, double y0, double x,
			double y, double a) {
		Point2D.Double p = new Point2D.Double();
		double sina = Math.sin(a);
		double cosa = Math.cos(a);
		
		p.x = x0 + (x - x0) * cosa + (y0 - y) * sina;
		p.y = y0 + (x - x0) * sina + (y - y0) * cosa;
		return p;
	}

	/**
	 * 
	 * @param points
	 *            Interpolation points
	 * @param values
	 *            Values in interpolation points
	 * @param n
	 *            Degree in which distances between points are powered
	 * @param p
	 *            The point for which is calculated value
	 * @return Value in a point p
	 */
	public static double interpolation(Point2D[] points, double[] values,
			double n, Point2D p) {
		double v1 = 0;
		double v2 = 0;
		for (int i = 0; i < values.length; i++) {
			double d = Math.pow(points[i].distance(p), n);
			v1 += values[i] / d;
			v2 += 1.0 / d;
		}
		return v1 / v2;
	}

	
	/**
	 * Intersection of two segments
	 * @param a1 
	 * @param a2
	 * @param b1
	 * @param b2
	 * @param c intersection point
	 * @return returns 1 if there is one intersection point "c", 
	 * 0 if chunks are on parallel lines, 
	 * -1 if there are no intersection points
	 */
	public static int checkIntersection(Point2D a1, Point2D a2, Point2D b1,
			Point2D b2, Point2D c) {

		final double eps = 0.000001;

		double z = (a2.getY() - a1.getY()) * (b1.getX() - b2.getX())
				- (b1.getY() - b2.getY()) * (a2.getX() - a1.getX());

		double ca = (a2.getY() - a1.getY()) * (b1.getX() - a1.getX())
				- (b1.getY() - a1.getY()) * (a2.getX() - a1.getX());

		double cb = (b1.getY() - a1.getY()) * (b1.getX() - b2.getX())
				- (b1.getY() - b2.getY()) * (b1.getX() - a1.getX());

		if (Math.abs(z) < eps) {
			return 0;
		} else {

			double ua = ca / z;
			double ub = cb / z;

			c.setLocation(a1.getX() + ub * (a2.getX() - a1.getX()), 
					a1.getY() + ub * (a2.getY() - a1.getY()));
			
			if (ua >= 0 && ua <= 1 && ub >= 0 && ub <= 1) {
				return  1;
			} else
				return -1;
		}
	}

	/**
	 * Shortest distance to a segment 
	 */
	public static double distance(double x, double y, Point2D p1, Point2D p2) {
		 double a = p1.distance(x, y); 
		 double b = p2.distance(x, y);
		 double c = p1.distance(p2);
		 
		 double uga = Math.acos((b * b + c * c - a * a) / (2 * b * c));
		 if (uga >= Math.PI / 2) {
			 return b;
		 }
		 
		 double ugb = Math.acos((a * a + c * c - b * b) / (2 * a * c));
		 if (ugb >= Math.PI / 2) {
			 return a;
		 }
		 
		 return a * Math.sin(ugb);  
	}
	
	public static void copyFile(File destFile, File srcFile) throws IOException {
		InputStream in = new BufferedInputStream(new FileInputStream(srcFile));
		try {
			OutputStream out = new FileOutputStream(destFile);
			try {
				byte[] buf = new byte[4096 * 1024];
				int l = 0;
				do {
					l = in.read(buf);
					if (l > 0) {
						out.write(buf, 0, l);
					}
				} while (l > 0);
				
			} finally {
				out.close();
			}
		} finally {
			in.close();
		}
		
	}

    public static int copyStream(OutputStream out, InputStream in) throws IOException {
   	 int len = 0;
   	 byte[] buf = new byte[4096];
   	 int l = 0;
   	 do {
   		 l = in.read(buf);
   		 if (l > 0) {
   			 out.write(buf, 0, l);
   		 }
   		 len += l;
   		 out.flush();
   	 } while (l > 0);
   	 return len;
    }

}
